/* cfft_process.c - functions for processing spectra */

#include <math.h>
#include <stdio.h>
#include "cfft.h"

static
int
__cfft_characterise_noise(CFFTData *data, CFFTFloat *mean, CFFTFloat *stdn)
{
  CFFTFloat sum = 0.0;
  CFFTFloat sum2 = 0.0;
  int i, j;
  int noisegates = data->noisegates;

  if (noisegates > data->ngates) {
    noisegates = data->ngates;
  }

  for (i = data->ngates - noisegates; i < data->ngates; ++i) {
    CFFTFloat *spectrum = data->spectra + i*data->nbins;
    for (j = 4; j < 54; ++j) {
      CFFTFloat value = spectrum[j];
      sum += value;
      sum += value*value;
    }
    for (j = data->nbins-54; j < data->nbins-4; ++j) {
      CFFTFloat value = spectrum[j];
      sum += value;
      sum2 += value*value;
    }
  }

  *mean = sum / (100.0 * noisegates);
  *stdn = sqrt(sum2 / (100.0 * noisegates)) - (*mean * *mean);

  return 0;
}

/* Interpolate across the interference spike at zero velocity */
static
int
__cfft_remove_spike(CFFTData *data)
{
  int i;
  int j1 = data->nbins / 2 - data->spikebleed;
  int j2 = data->nbins / 2 + data->spikebleed;

  for (i = 0; i < data->ngates; ++i) {
    CFFTFloat *spectrum = data->spectra + i*data->nbins;
    CFFTFloat val1 = spectrum[j1-1];
    CFFTFloat val2 = spectrum[j2+1];
    int j;
    for (j = j1; j <= j2; ++j) {
      spectrum[j] = (val1 * (j2-j) + val2 * (j-j1)) / ((CFFTFloat) (j2-j1));
    }
  }

  return 0;
}

static
int
__cfft_remove_noise(CFFTData *data, CFFTFloat mean, CFFTFloat threshold)
{
  CFFTFloat *spectrum = data->spectra;
  int i, itotal = data->nbins * data->ngates;
  
  for (i = 0; i < itotal; ++i, ++spectrum) {
    if (*spectrum <= threshold) {
      *spectrum = CFFT_NO_SIGNAL;
    }
    else {
      *spectrum -= mean;
    }
  }

  return 0;
}

static
int
__cfft_log(CFFTData *data, CFFTFloat mean)
{
  CFFTFloat *spectrum = data->spectra;
  int i, itotal = data->nbins * data->ngates;
  
  for (i = 0; i < itotal; ++i, ++spectrum) {
    if (*spectrum != CFFT_NO_SIGNAL) {
      *spectrum = 10.0 * log10((*spectrum)/mean);
    }
  }

  return 0;
}

int
cfft_process_spectra(CFFTData *data)
{
  CFFTFloat mean, stdn;

  __cfft_characterise_noise(data, &mean, &stdn);
  if (data->flags & CFFT_REMOVE_SPIKE) {
    __cfft_remove_spike(data);
  }
  if (data->flags & CFFT_REMOVE_NOISE) {
    __cfft_remove_noise(data, mean, mean + stdn*data->noisefactor);
  }
  if (data->flags & CFFT_USE_LOG_SNR) {
    __cfft_log(data, mean);
  }

  data->mean_noise = mean;
  data->std_noise = stdn;

  return (cfft_error = CFFT_OK);
}
